#import "HelloVesSceneFactory.h"

#include <vtkSmartPointer.h>
#include <vtkSphereSource.h>
#include <vtkCubeSource.h>
#include <vtkElevationFilter.h>
#include <vtkPolyData.h>
#include <vtkGlyph3D.h>
#include <vtkPolyDataNormals.h>

#include <vesMapper.h>
#include <vesActor.h>
#include <vesBlend.h>
#include <vesDepth.h>
#include <vesImage.h>
#include <vesTexture.h>

#import "mvesView.h"
#import "mvesShaderPrograms.h"
#import "mvesPolyDataMapper.h"
#import "mvesRenderer.h"
#import "mvesImageRepresentation.h"
#import "mvesRectangleSource.h"
#import "mvesMaterials.h"

#import "HelloVesScene.h"
#import "HelloVesSceneTableViewController.h"


#pragma mark - balls scene

@interface BallsScene : NSObject <HelloVesScene> {
    int dir;
    vtkSmartPointer<vtkSphereSource> sphereSource;
    vtkSmartPointer<vtkElevationFilter> elevationFilter;
    vtkSmartPointer<vtkPolyData> glyphData;
    vtkSmartPointer<vtkPoints>pts;
    vtkSmartPointer<vtkGlyph3D> glyphFilter;
    vtkSmartPointer<vtkPolyDataNormals> normals;
    vesSharedPtr<mvesPolyDataMapper> mapper;
    vesSharedPtr<vesMaterial> material;
    vesSharedPtr<vesActor> actor;
}

@end


@implementation BallsScene

- (id)init
{
    self = [super init];
    if(self != nil) {
        dir = -1;
        sphereSource = vtkSmartPointer<vtkSphereSource>::New();
        elevationFilter = vtkSmartPointer<vtkElevationFilter>::New();
        glyphFilter = vtkGlyph3D::New();
        pts = vtkPoints::New();
        glyphData = vtkPolyData::New();
        normals = vtkPolyDataNormals::New();
        mapper = vesSharedPtr<mvesPolyDataMapper>(new mvesPolyDataMapper());
        material = vesSharedPtr<vesMaterial>(new vesMaterial());
        actor = vesSharedPtr<vesActor>(new vesActor());
        
        // sphere source
        sphereSource->SetCenter(0, 0, 0);
        sphereSource->SetRadius(1);
        sphereSource->SetPhiResolution(15);
        sphereSource->SetThetaResolution(15);
        
        // points
        glyphData->SetPoints(pts);
        for(int i = 0; i < 200; i++) {
            double x1 = (10.0 * rand() / (RAND_MAX + 1.0));
            double x2 = (10.0 * rand() / (RAND_MAX + 1.0));
            double x3 = (10.0 * rand() / (RAND_MAX + 1.0));
            pts->InsertNextPoint(x1, x2, x3);
        }
        
        // glyph
        glyphFilter->SetSourceConnection(sphereSource->GetOutputPort());
        glyphFilter->SetInputData(glyphData);
        
        // elevation filter
        elevationFilter->SetInputConnection(glyphFilter->GetOutputPort());
        elevationFilter->SetLowPoint(0, 0, 0);
        elevationFilter->SetHighPoint(0, 0, 10);
        
        // normals
        normals->SetInputConnection(elevationFilter->GetOutputPort());
        
        // mapper
        mapper->SetInputConnection(normals->GetOutputPort());
        mapper->getLookupTable()->SetRange(0, 1);
        
        // material
        material->addAttribute(mvesShaderPrograms::gouraudShaderProgram());
        material->addAttribute(vesSharedPtr<vesBlend>(new vesBlend()));
        material->addAttribute(vesSharedPtr<vesDepth>(new vesDepth()));
        
        // actor
        actor->setMapper(mapper);
        actor->setMaterial(material);
    }
    return self;
}

- (void)update
{
    double s = sphereSource->GetRadius() + dir * 0.01;
    
    sphereSource->SetRadius(s);
    if(s < 0.4 || s >= 1) {
        dir *= -1;
    }
}

- (void)addActorsToRenderer:(vesSharedPtr<mvesRenderer>)r
{
    r->addActor(actor);
}

- (void)removeActorsFromRenderer:(vesSharedPtr<mvesRenderer>)r
{
    r->removeActor(actor);
}

- (int)cellCount
{
    return glyphFilter->GetOutput()->GetNumberOfCells();
}

@end


#pragma mark - textured cube scene

@interface CubeScene : NSObject <HelloVesScene> {
    vtkSmartPointer<vtkCubeSource> cubeSource;
    vesSharedPtr<mvesPolyDataMapper> mapper;
    vesSharedPtr<vesImage> image;
    vesSharedPtr<vesTexture> texture;
    vesSharedPtr<vesMaterial> material;
    vesSharedPtr<vesActor> actor;
}

@end


@implementation CubeScene

- (id)init
{
    self = [super init];
    if(self != nil) {
        cubeSource = vtkSmartPointer<vtkCubeSource>::New();
        mapper = vesSharedPtr<mvesPolyDataMapper>(new mvesPolyDataMapper());
        image = vesSharedPtr<vesImage>(new vesImage());
        texture = vesSharedPtr<vesTexture>(new vesTexture());
        material = vesSharedPtr<vesMaterial>(new vesMaterial());
        actor = vesSharedPtr<vesActor>(new vesActor());
        
        // image
        int n = 20;
        int size = n * n * 3;
        unsigned char imageData[size];
        for (int i = 0; i < size; i++) {
            imageData[i] = rand() % 255;
        }
        image->setWidth(n);
        image->setHeight(n);
        image->setPixelDataType(vesColorDataType::UnsignedByte);
        image->setPixelFormat(vesColorDataType::RGB);
        image->setData(imageData, size * sizeof(unsigned char));
        
        // texture
        texture->setImage(image);
        
        // mapper
        mapper->SetInputConnection(cubeSource->GetOutputPort());
        
        // material
        material->addAttribute(mvesShaderPrograms::gouraudTextureShaderProgram());
        material->addAttribute(vesSharedPtr<vesBlend>(new vesBlend()));
        material->addAttribute(vesSharedPtr<vesDepth>(new vesDepth()));
        material->addAttribute(vesSharedPtr<vesTexture>(texture));
        
        // actor
        actor->setMapper(mapper);
        actor->setMaterial(material);
    }
    return self;
}

- (void)update
{
}

- (void)addActorsToRenderer:(vesSharedPtr<mvesRenderer>)r
{
    r->addActor(actor);
}

- (void)removeActorsFromRenderer:(vesSharedPtr<mvesRenderer>)r
{
    r->removeActor(actor);
}

- (int)cellCount
{
    return cubeSource->GetOutput()->GetNumberOfCells();
}

@end


#pragma mark - image scene

@interface ImageScene : NSObject <HelloVesScene>

@property (nonatomic) int cnt;
@property (nonatomic) mvesImageRepresentation *imageRep;

@end


@implementation ImageScene

- (id)init
{
    self = [super init];
    if(self != nil) {
        double p[3] = {0, 0, 0};
        double s1[3] = {5, 1, 0};
        double v[3] = {0, 3, 1};
        UIImage *img = [UIImage imageNamed:@"image.jpg"];
        
        _imageRep = [[mvesImageRepresentation alloc] initWithImage:img at:p width:2];        
        [_imageRep setOrientation:s1 v:v];
    }
    return self;
}

- (void)update
{
    self.imageRep.opacity = 0.5 + 0.25 * (sin(0.1 * self.cnt) + 1);
    [self.imageRep rotateByAngle:0.25];
    self.cnt++;
}

- (void)addActorsToRenderer:(vesSharedPtr<mvesRenderer>)r
{
    [self.imageRep addToRenderer:r];
}

- (void)removeActorsFromRenderer:(vesSharedPtr<mvesRenderer>)r
{
    [self.imageRep removeFromRenderer:r];
}

- (int)cellCount
{
    return 1;
}

@end


#pragma mark - factory

static HelloVesSceneFactory* _sharedInstance = nil;

@implementation HelloVesSceneFactory

+ (HelloVesSceneFactory *)sharedInstance
{
    @synchronized(self) {
        _sharedInstance = [[HelloVesSceneFactory alloc] init];
    }
    return _sharedInstance;
}

- (NSArray *)scenes
{
    return @[@"Balls", @"Textured Cube", @"Image Representation"];
}

- (id<HelloVesScene>)createScene:(NSString *)name
{
    int idx = [self.scenes indexOfObject:name];
    
    if(idx == 0){
        return [[BallsScene alloc] init];
    }
    if(idx == 1) {
        return [[CubeScene alloc] init];
    }
    return [[ImageScene alloc] init];
}

@end
