#import "mvesImageRepresentation.h"

#include <vtkSmartPointer.h>
#include <vtkPolyData.h>
#include <vtkPointData.h>
#include <vtkFloatArray.h>

#include <vesMapper.h>
#include <vesActor.h>
#include <vesBlend.h>
#include <vesDepth.h>
#include <vesImage.h>
#include <vesTexture.h>

#import "mvesView.h"
#import "mvesRenderer.h"
#import "mvesMaterials.h"
#import "mvesShaderPrograms.h"
#import "mvesPolyDataMapper.h"
#import "mvesRectangleSource.h"


@interface mvesImageRepresentation() {
    vtkSmartPointer<mvesRectangleSource> source;
    vesSharedPtr<mvesPolyDataMapper> mapper;
    vesSharedPtr<vesImage> image;
    vesSharedPtr<vesTexture> texture;
    vesSharedPtr<vesMaterial> material;
    vesSharedPtr<vesActor> actor;
}
@end


@implementation mvesImageRepresentation

- (id)initWithImage:(UIImage *)uiImage at:(double [3])c width:(double)w
{
    self = [self init];
    if(self != nil) {
        [self setImage:uiImage];
        [self setCenter:c];
        self.width = w;
    }
    return self;
}

- (id)init
{
    self = [super init];
    if(self != nil) {
        source = vtkSmartPointer<mvesRectangleSource>::New();
        mapper = vesSharedPtr<mvesPolyDataMapper>(new mvesPolyDataMapper());
        image = vesSharedPtr<vesImage>(new vesImage());
        texture = vesSharedPtr<vesTexture>(new vesTexture());
        material = vesSharedPtr<vesMaterial>(new vesMaterial());
        actor = vesSharedPtr<vesActor>(new vesActor());
        
        // mapper
        mapper->SetInputConnection(source->GetOutputPort());
        mapper->setColor(1.0, 1.0, 1.0, 1.0);
        
        // material
        material = mvesMaterials::gouraudTextureMaterial(texture);
        
        // actor
        actor->setMapper(mapper);
        actor->setMaterial(material);
    }
    return self;
}

- (double)x
{
    return source->GetCenter()[0];
}

- (void)setX:(double)x
{
    double p[3];
    source->GetCenter(p);
    p[0] = x;
    source->SetCenter(p);
}

- (double)y
{
    return source->GetCenter()[1];
}

- (void)setY:(double)y
{
    double p[3];
    source->GetCenter(p);
    p[1] = y;
    source->SetCenter(p);
}

- (double)z
{
    return source->GetCenter()[2];
}

- (void)setZ:(double)z
{
    double p[3];
    source->GetCenter(p);
    p[2] = z;
    source->SetCenter(p);
}

- (double)width
{
    return source->GetWidth();
}

- (void)setWidth:(double)w
{
    double iw = image->width();
    double ih = image->height();

    source->SetWidth(w);
    source->SetHeight(ih / iw * w);
}

- (float)opacity
{
    return mapper->color()[3];
}

- (void)setOpacity:(float)o
{
    mapper->color()[3] = o;
}

- (void)addToRenderer:(vesSharedPtr<mvesRenderer>)r
{
    r->addActor(actor);
}

- (void)removeFromRenderer:(vesSharedPtr<mvesRenderer>)r
{
    r->removeActor(actor);
}

- (void)setImage:(UIImage*)img
{
    if (img != nil) {
        int nw = img.size.width;
        int nh = img.size.height;
        CGImageRef cgImage = [img CGImage];
        CFDataRef data = CGDataProviderCopyData(CGImageGetDataProvider(cgImage));
        const unsigned char * buffer = CFDataGetBytePtr(data);
        int size = CFDataGetLength(data);
        int nBytes = size / (nw * nh); // number of bytes per pixel
        vesColorDataType::PixelFormat pixelFormat;
        
        // educated guess for pixel format
        switch (nBytes) {
            case 1:
                pixelFormat = vesColorDataType::Luminance;
                break;
            case 2:
                pixelFormat = vesColorDataType::LuminanceAlpha;
                break;
            case 3:
                pixelFormat = vesColorDataType::RGB;
                break;
            default:
                pixelFormat = vesColorDataType::RGBA;
                break;
        }
        
        image->setWidth(nw);
        image->setHeight(nh);
        image->setPixelFormat(pixelFormat);
        image->setPixelDataType(vesColorDataType::UnsignedByte);
        image->setData(buffer, size * sizeof(unsigned char));
        source->SetHeight(self.width * ((double)nh) / ((double)nw));
        texture->setImage(image);
        
        CFRelease(data);
    } else {
        image->setWidth(1);
        image->setHeight(1);
        image->setPixelDataType(vesColorDataType::UnsignedByte);
        image->setPixelFormat(vesColorDataType::RGBA);
        const unsigned char buffer[4] = {255.0, 255.0, 255.0, 255.0};
        image->setData(buffer, 4);
        texture->setImage(image);
    }
}

- (void)setCenter:(double [3])p
{
    source->SetCenter(p);
}

- (void)scale:(double)s
{
    source->scale(s);
}

- (void)rotateByAngle:(double)a
{
    source->rotate(a);
}

- (void)rotateByAngle:(double)a round:(double[3])axis
{
    source->rotate(a, axis);
}

- (void)setOrientation:(double[3])s1 v:(double[3])v
{
    source->setOrientation(s1, v);
}

- (void)translate:(double[3])p
{
    source->translate(p[0], p[1], p[2]);
}


@end
