#include "mvesRectangleSource.h"

#include "vtkMath.h"
#include "vtkPoints.h"
#include "vtkPolyData.h"
#include "vtkCellArray.h"
#include "vtkTransform.h"
#include "vtkPointData.h"
#include "vtkDoubleArray.h"
#include "vtkInformation.h"
#include "vtkSmartPointer.h"
#include "vtkObjectFactory.h"
#include "vtkInformationVector.h"
#include "vtkStreamingDemandDrivenPipeline.h"

#include <math.h>


vtkStandardNewMacro(mvesRectangleSource);


void mvesRectangleSource::SetWidth(double w) {
    vtkMath::MultiplyScalar(this->S1, 0.5 * w / this->GetWidth());
    Modified();
}

double mvesRectangleSource::GetWidth() {
    return 2 * vtkMath::Norm(this->S1, 3);
}

void mvesRectangleSource::SetHeight(double h) {
    vtkMath::MultiplyScalar(this->S2, 0.5 * h / this->GetHeight());
    Modified();
}

double mvesRectangleSource::GetHeight() {
    return 2 * vtkMath::Norm(this->S2, 3);
}

void mvesRectangleSource::rotate(double alpha) {
    double n[3];
    
    vtkMath::Cross(this->S1, this->S2, n);
    rotate(alpha, n);
}

void mvesRectangleSource::rotate(double alpha, double axis[3]) {
    vtkSmartPointer<vtkTransform> t = vtkTransform::New();
    
    t->RotateWXYZ(alpha, axis);
    t->TransformPoint(this->S1, this->S1);
    t->TransformPoint(this->S2, this->S2);
    Modified();
}

void mvesRectangleSource::setOrientation(double s1[3], double v[3]) {
    double n[3];
    double w = GetWidth();
    double h = GetHeight();
    
    this->S1[0] = s1[0];
    this->S1[1] = s1[1];
    this->S1[2] = s1[2];
    vtkMath::Cross(s1, v, n);
    vtkMath::Cross(n, s1, this->S2);
    SetWidth(w);
    SetHeight(h);
    Modified();
}

void mvesRectangleSource::scale(double s) {
    vtkMath::MultiplyScalar(this->S1, s);
    vtkMath::MultiplyScalar(this->S2, s);
}

void mvesRectangleSource::translate(double dx, double dy, double dz) {
    this->Center[0] += dx;
    this->Center[1] += dy;
    this->Center[2] += dz;
    Modified();
}

mvesRectangleSource::mvesRectangleSource() {
    SetNumberOfInputPorts(0);
    SetNumberOfOutputPorts(1);
    this->Center[0] = 0;
    this->Center[1] = 0;
    this->Center[2] = 0;
    this->S1[0] = 1;
    this->S1[1] = 0;
    this->S1[2] = 0;
    this->S2[0] = 0;
    this->S2[1] = 1;
    this->S2[2] = 0;
}

int mvesRectangleSource::RequestData(vtkInformation *vtkNotUsed(request), vtkInformationVector **vtkNotUsed(inputVector), vtkInformationVector *outputVector) {
    double normal[3];
    vtkInformation *outInfo0 = outputVector->GetInformationObject(0);
    vtkPolyData *outPD = vtkPolyData::SafeDownCast(outInfo0->Get(vtkDataObject::DATA_OBJECT()));
    vtkSmartPointer<vtkPoints> points = vtkSmartPointer<vtkPoints>::New();
    vtkSmartPointer<vtkCellArray> polys = vtkSmartPointer<vtkCellArray>::New();
    vtkSmartPointer<vtkDoubleArray> tCoords = vtkSmartPointer<vtkDoubleArray>::New();
    vtkSmartPointer<vtkDoubleArray> normals = vtkSmartPointer<vtkDoubleArray>::New();
    
    tCoords->SetNumberOfComponents(2);
    tCoords->SetName("TextureCoordinates");
    normals->SetNumberOfComponents(3);
    normals->SetName("Normals");
    outPD->SetPoints(points);
    outPD->SetPolys(polys);
    outPD->GetPointData()->SetTCoords(tCoords);
    outPD->GetPointData()->SetNormals(normals);

    polys->InsertNextCell(4);

    points->InsertNextPoint(this->Center[0] - this->S1[0] - this->S2[0], this->Center[1] - this->S1[1] - this->S2[1], this->Center[2] - this->S1[2] - this->S2[2]);
    points->InsertNextPoint(this->Center[0] + this->S1[0] - this->S2[0], this->Center[1] + this->S1[1] - this->S2[1], this->Center[2] + this->S1[2] - this->S2[2]);
    points->InsertNextPoint(this->Center[0] + this->S1[0] + this->S2[0], this->Center[1] + this->S1[1] + this->S2[1], this->Center[2] + this->S1[2] + this->S2[2]);
    points->InsertNextPoint(this->Center[0] - this->S1[0] + this->S2[0], this->Center[1] - this->S1[1] + this->S2[1], this->Center[2] - this->S1[2] + this->S2[2]);

    tCoords->InsertNextTuple2(0, 1);
    tCoords->InsertNextTuple2(1, 1);
    tCoords->InsertNextTuple2(1, 0);
    tCoords->InsertNextTuple2(0, 0);

    polys->InsertCellPoint(0);
    polys->InsertCellPoint(1);
    polys->InsertCellPoint(2);
    polys->InsertCellPoint(3);

    vtkMath::Cross(this->S1, this->S2, normal);
    normals->InsertNextTupleValue(normal);
    normals->InsertNextTupleValue(normal);
    normals->InsertNextTupleValue(normal);
    normals->InsertNextTupleValue(normal);
    
    return 1;
}
